/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.collections.adapters;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

/**
 * Fully mutable adaper from List&gt;E1&lt; to List&gt;E2&lt; 
 * @author borisman
 *
 * @param <E1>
 * @param <E2>
 */
public class TwoWayListAdapter<E1, E2> implements List<E2>{
	
	private final TwoWayAdapter<E1, E2> adapter;
	private final List<E1> backend;
	
	public TwoWayListAdapter(final TwoWayAdapter<E1, E2> adapter, final List<E1> backend) {
		this.adapter = adapter;
		this.backend = backend;
	}

	@Override public boolean add(final E2 e) {
		return backend.add(adapter.forward(e));
	}

	@Override public void add(final int index, final E2 element) {
		backend.add(index, adapter.forward(element));
	}

	@Override public boolean addAll(final Collection<? extends E2> c) {
		boolean res = false;
		for (final E2 e : c) {
			res = res & add(e);
		}
		return res;
	}

	@Override public boolean addAll(int index, final Collection<? extends E2> c) {
		for (final E2 e : c) {
			add(index, e);
			index++;
		}
		return true;
	}

	@Override public void clear() {
		backend.clear();
	}

	@SuppressWarnings("unchecked")
	@Override public boolean contains(final Object o) {
		if (!adapter.isE2(o)) return false;
		final E2 e2 = (E2) o;
		return backend.contains(adapter.forward(e2));
	}

	@Override public boolean containsAll(final Collection<?> c) {
		for (final Object o : c) {
			if (!contains(o)) return false;
		}
		return true;
	}

	@Override public E2 get(final int index) {
		return adapter.backward(backend.get(index));
	}

	@SuppressWarnings("unchecked")
	@Override public int indexOf(final Object o) {
		if (!adapter.isE2(o)) return -1;
		final E2 e2 = (E2) o;
		return backend.indexOf(adapter.forward(e2));
	}

	@Override public boolean isEmpty() {
		return backend.isEmpty();
	}

	@Override public Iterator<E2> iterator() {
		return new TwoWayListIteratorAdapter<E1, E2>(adapter, backend.listIterator());
	}

	@SuppressWarnings("unchecked")
	@Override public int lastIndexOf(final Object o) {
		if (!adapter.isE2(o)) return -1;
		final E2 e2 = (E2) o;
		return backend.lastIndexOf(adapter.forward(e2));
	}

	@Override public ListIterator<E2> listIterator() {
		return new TwoWayListIteratorAdapter<E1, E2>(adapter, backend.listIterator());
	}

	@Override public ListIterator<E2> listIterator(final int index) {
		return new TwoWayListIteratorAdapter<E1, E2>(adapter, backend.listIterator(index));
	}

	@SuppressWarnings("unchecked")
	@Override public boolean remove(final Object o) {
		if (!adapter.isE2(o)) return false;
		final E2 e2 = (E2) o;
		return backend.remove(adapter.forward(e2));
	}

	@Override public E2 remove(final int index) {
		return adapter.backward(backend.remove(index));
	}

	@Override public boolean removeAll(final Collection<?> c) {
		return backend.removeAll(recastCollection(c));
	}
	
	@Override public boolean retainAll(final Collection<?> c) {
		return backend.retainAll(recastCollection(c));
	}

	@Override public E2 set(final int index, final E2 element) {
		final E1 e1 = adapter.forward(element);
		final E1 res = backend.set(index, e1); 
		return adapter.backward(res);
	}

	@Override public int size() {
		return backend.size();
	}

	@Override public List<E2> subList(int fromIndex, int toIndex) {
		throw new UnsupportedOperationException();
	}

	@Override public Object[] toArray() {
		final Object [] res = backend.toArray();
		remapArrayFromE1toE2(res);
		return res;
	}

	@SuppressWarnings("unchecked")
	@Override public <T> T[] toArray(T[] a) {
		final Object [] res = backend.toArray(a);
		remapArrayFromE1toE2(res);
		return (T[]) res;
	}
	
	@SuppressWarnings("unchecked")
	private void remapArrayFromE1toE2(final Object[] array) {
		for (int i=0; i<backend.size(); i++) {
			final E1 e1 = (E1) array [i];
			final E2 e2 = adapter.backward(e1);
			array [i] = e2;
		}
	}

	
	@SuppressWarnings("unchecked")
	private List<E1> recastCollection(final Collection<?> c) {
		final List<E1> coll = new ArrayList<E1> (c.size());
		for (final Object o : c) {
			if (adapter.isE2(o)) continue;
			final E2 e2 = (E2) o;
			final E1 e1 = adapter.forward(e2);
			coll.add(e1);
		}
		return coll;
	}

}
