/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.collections.bytearray;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.Collection;

import org.builtu.exceptions.FatalError;


/**
 * Very efficient implementation of the List&lt;byte&gt; interface backed by 
 * simple byte[] array. Has several specialized methods that make this class 
 * even more efficient: 
 *   {@link ByteArray#byte_add(byte)}
 *   {@link ByteArray#byte_add(byte[])}
 *   {@link ByteArray#byte_add(int, byte)}
 *   {@link ByteArray#byte_add(int, byte[], int, int)}
 *   {@link ByteArray#byte_contains(byte)}
 *   {@link ByteArray#byte_get(int)}
 *   {@link ByteArray#byte_indexOf(byte)}
 *   {@link ByteArray#byte_lastIndexOf(byte)}
 *   {@link ByteArray#byte_remove(byte)}
 *   {@link ByteArray#byte_removeAll(byte[])}
 *   {@link ByteArray#byte_set(int, byte)}
 *   {@link ByteArray#byte_set(int, byte[])}
 *   {@link ByteArray#byte_set(int, byte[], int, int)}
 * @author borisman
 *
 */
public class ByteArray extends FixedByteArray
{
  public ByteArray() {
    this(16);
  }
  
  public ByteArray(final int initialSize) {
    super(new byte[initialSize]);
    this.size=0;
  }
  
  @Override public boolean add(final Byte b)
  {
    byte_add(b);
    return true;
  }
  
  public void byte_add(final byte b) {
    final int newSize = grow(size+1);
    arr[size] = b;
    size = newSize;
  }

  @Override public void add(final int idx, final Byte b)
  {
    byte_add(idx, b);
  }
  
  
  public boolean byte_add(final int idx, final byte b) {
    if (idx<0 || idx>size) 
      throw new IndexOutOfBoundsException("idx="+idx+", size="+size);
    final int newSize = grow(size+1);
    if (idx<size) System.arraycopy(arr, idx, arr, idx+1, size-idx);
    arr[idx] = b;
    size = newSize;
    return true;
  }  
  
  public boolean byte_add(final byte [] bytes) {
    return byte_add(size, bytes, 0, bytes.length);
  }
  
  public boolean byte_add(
      final int idx,
      final byte[] bytes, 
      final int srcStartIdx, 
      final int srcEndIdx
  ) {
    if (idx<0 || idx>size) 
      throw new IndexOutOfBoundsException("idx="+idx+", size="+size);
    
    final int toAdd = srcEndIdx - srcStartIdx;
    if (toAdd<0) throw new IllegalArgumentException("srcEndIdx="+srcEndIdx+
        "<srcStartIdx="+srcStartIdx);
    if (toAdd == 0) return false;
    final int newSize = prepareToAdd(idx, toAdd);
    System.arraycopy(bytes, srcStartIdx, arr, idx, toAdd);
    size = newSize;
    return true;
  }

  @Override public boolean addAll(final Collection<? extends Byte> coll)
  {
    return addAll(size, coll);
  }

  @Override public boolean addAll(
      final int idx, 
      final Collection<? extends Byte> coll)
  {
    if (idx<0 || idx>size) 
      throw new IndexOutOfBoundsException("idx="+idx+", size="+size);

    final int toAdd = coll.size();
    if (toAdd == 0) return false;
    final int newSize = prepareToAdd(idx, toAdd);
    int i=idx;
    for(final Byte b : coll) arr[i++] = b;
    assert i-idx == toAdd;
    size = newSize;
    return true;
  }

  private int prepareToAdd(
      final int idx, 
      final int toAdd
  ) {
    assert toAdd>=0;
    if (toAdd==0) return size;
    final int newSize = grow(size+toAdd);
    if(idx<size) System.arraycopy(arr, idx, arr, idx+toAdd, size-idx);
    return newSize;
  }

  @Override public void clear()
  {
    size = 0;
  }

  @Override public Byte remove(final int idx)
  {
    checkIdx(idx);
    final byte res = arr[idx];
    if (idx<size-1) System.arraycopy(arr, idx+1, arr, idx, size-idx-1);
    size--;
    return res;
  }

  @Override public boolean remove(final Object o)
  {
    return byte_remove((Byte) o);
  }
  
  public boolean byte_remove(final byte b) {
    final int idx = byte_indexOf(b);
    if (idx<0) return false; 
    remove(idx);
    return true;
  }
  
  @Override public boolean removeAll(final Collection<?> coll)
  {
    boolean res=false;
    for (final Object o : coll) {
      int idx = indexOf(o);
      while (idx>=0) {
        remove(idx);
        res = true;
        idx = indexOf(o);
      }
    }
    return res;
  }
  
  public boolean byte_removeAll(final byte [] bytes) {
    boolean res=false;
    for (final Object o : bytes) {
      int idx = indexOf(o);
      while (idx>=0) {
        remove(idx);
        res = true;
        idx = indexOf(o);
      }
    }
    return res;
  }

  @Override public boolean retainAll(final Collection<?> coll)
  {
    int i=0;
    boolean res = false;
    while (i<size) {
      if (coll.contains(arr[i])) i++;
      else {
        remove(i);
        res = true;
      }
    }
    return res;
  }
  
  private int grow(final int newSize)
  {
    if (newSize>arr.length) {
      final byte [] newArr = new byte[newSize << 1];
      System.arraycopy(arr, 0, newArr, 0, size);
      this.arr = newArr;
    }
    return newSize;
  }

  public void readAll(final InputStream in) throws IOException
  {
    while(true) {
      if (size == arr.length) grow(size << 1);
      int read = in.read(arr, size, arr.length-size);
      if (read<0) break;
      size += read;
    }
  }
  
  public String asUTF8() {
    try {
      return new String(arr, 0, size, "UTF-8");
    } catch (final UnsupportedEncodingException e) {
      throw new FatalError(e);
    }
  }
  
  private class ByteArrayOutputStream extends OutputStream {
    
    @Override public void write(
        final byte[] bytes, 
        final int off, 
        final int len
    ) throws IOException {
      ByteArray.this.byte_add(size, bytes, off, off+len);
    }

    @Override public void write(final byte[] bytes) throws IOException
    {
      ByteArray.this.byte_add(bytes);
    }

    @Override public void write(final int b) throws IOException
    {
      ByteArray.this.byte_add((byte)b);
    }
    
  }

  public OutputStream asOutputStream()
  {
    return new ByteArrayOutputStream();
  }

  
}
