/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.collections.bytearray;

import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.NoSuchElementException;

/**
 * Very efficiently implements parts of List&lt;Byte&gt; interface on fixed byte
 * array. NOTE! METHODS, THAT MAY CHANGE LIST SIZE ARE NOT SUPPORTED!!!
 * In particular, these methods will throw UnsupportedOperationException:
 *    {@link FixedByteArray#add(Byte)}
 *    {@link FixedByteArray#add(int, Byte)}
 *    {@link FixedByteArray#addAll(Collection)}
 *    {@link FixedByteArray#addAll(int, Collection)}
 *    {@link FixedByteArray#clear()}
 *    {@link FixedByteArray#remove(int)}
 *    {@link FixedByteArray#remove(Object)}
 *    {@link FixedByteArray#removeAll(Collection)}
 *    {@link FixedByteArray#retainAll(Collection)}
 * This mens also that iterators will not support modifying functions as well:
 *    {@link ByteArrayIterator#add(Byte)}
 *    {@link ByteArrayIterator#remove()}
 * Using these methods will make code even more efficient:
 *    {@link FixedByteArray#byte_contains(byte)}
 *    {@link FixedByteArray#byte_get(int)}
 *    {@link FixedByteArray#byte_indexOf(byte)}
 *    {@link FixedByteArray#byte_lastIndexOf(byte)}
 *    {@link FixedByteArray#byte_set(int, byte)}
 *    {@link FixedByteArray#byte_set(int, byte[])}
 *    {@link FixedByteArray#byte_set(int, byte[], int, int)}
 * @author borisman
 */
public class FixedByteArray implements List<Byte>
{
  protected byte[] arr;
  private final int start;
  protected int size;
  
  public FixedByteArray(final byte[] arr) {
    this(arr, 0, arr.length);
  }
  
  public FixedByteArray(final byte[] arr, final int start, final int end) {
    if (start<0) throw new IndexOutOfBoundsException("start="+start+"<0");
    if (end<0) throw new IndexOutOfBoundsException("end="+end+"<0");
    if (start>arr.length) throw new 
        IndexOutOfBoundsException("start="+start+">arr.length="+arr.length);
    if (end>arr.length) throw new 
        IndexOutOfBoundsException("end="+end+">arr.length="+arr.length);
    if (end<start) throw new 
        IndexOutOfBoundsException("start="+start+">end="+end);
    this.arr = arr;
    this.start = start;
    this.size = end - start;
  }
  
  @Override public boolean equals(final Object o)
  {
    if (o == this) return true;
    if (o instanceof FixedByteArray) {
      final FixedByteArray rhs = (FixedByteArray) o;
      if (this.size != rhs.size) return false;
      return isEqualsToArray(rhs.arr, rhs.start);
    }
    if (!(o instanceof List<?>)) return false;
    final List<?> rhs = (List<?>) o;
    if (size!=rhs.size()) return false;
    for (int i=0; i<size; i++) {
      final Byte b = get(i);
      if (!b.equals(rhs.get(i))) return false;
    }
    return true;
  }
  
  private boolean isEqualsToArray(final byte [] rhs, final int rhs_start) {
    for (int i=0; i<size; i++) 
      if (arr[start+i]!=rhs[rhs_start+i]) return false;
    return true;
  }

  /**
   * Not supported
   */
  @Override public boolean add(final Byte b)
  {
    throw new UnsupportedOperationException();
  }

  /**
   * Not supported
   */
  @Override public void add(final int idx, final Byte b)
  {
    throw new UnsupportedOperationException();
  }

  /**
   * Not supported
   */
  @Override public boolean addAll(final Collection<? extends Byte> coll)
  {
    throw new UnsupportedOperationException();
  }

  /**
   * Not supported
   */
  @Override public boolean addAll(
      final int idx, 
      final Collection<? extends Byte> coll
  ) {
    throw new UnsupportedOperationException();
  }

  /**
   * Not supported
   */
  @Override public void clear()
  {
    throw new UnsupportedOperationException();
  }

  @Override public boolean contains(final Object o)
  {
    // Since Byte is final class, we can make this method very efficient
    if (o == null) return false;
    if (!(o instanceof Byte)) return false;
    return byte_contains((Byte) o);
  }
  
  public boolean byte_contains (final byte b) {
    return byte_indexOf(b)>=0;
  }

  @Override public boolean containsAll(final Collection<?> coll)
  {
    for (final Object o : coll) if (!contains(o)) return false;
    return true;
  }

  private class ByteArrayIterator implements ListIterator<Byte> 
  {
    private int pos, previousIdx=-1;
    
    public ByteArrayIterator(final int idx) 
    {
      this.pos = idx;
    }
    
    @Override public void add(final Byte b)
    {
      FixedByteArray.this.add(b);
    }

    @Override public boolean hasNext()
    {
      return pos-start<size;
    }

    @Override public boolean hasPrevious()
    {
      return pos>start;
    }

    @Override public Byte next()
    {
      if (!hasNext()) throw new NoSuchElementException();
      previousIdx = pos++;
      return arr[previousIdx];
    }

    @Override public int nextIndex()
    {
      if (!hasNext()) throw new NoSuchElementException();
      return pos-start;
    }

    @Override public Byte previous()
    {
      if (!hasPrevious()) throw new NoSuchElementException();
      previousIdx = --pos;
      return arr[previousIdx];
    }

    @Override public int previousIndex()
    {
      if (!hasPrevious()) throw new NoSuchElementException();
      return pos-1-start;
    }

    @Override public void remove()
    {
      if (previousIdx<0) throw new IllegalStateException();
      FixedByteArray.this.remove(previousIdx);
      pos = previousIdx;
      previousIdx=-1;
    }

    @Override public void set(final Byte e)
    {
      if (!hasNext()) throw new NoSuchElementException();
      FixedByteArray.this.set(pos, e);
    }
  }
  
  @Override public Byte get(final int idx)
  {
    return byte_get(idx);
  }

  public byte byte_get(final int idx)
  {
    checkIdx(idx);
    return arr[start+idx];
  }
  
  @Override public int indexOf(final Object o)
  {
    if (!(o instanceof Byte)) return -1;
    return byte_indexOf((Byte) o);
  }
  
  public int byte_indexOf(final byte b) {
    // Since Byte is final, make this little bit more efficient
    // Could be done even more efficiently by preallocating array of 256 
    // elements
    for (int i=0; i<size; i++) if(arr[start+i]==b) return i;
    return -1;
  }

  @Override public boolean isEmpty()
  {
    return size==0;
  }

  @Override public Iterator<Byte> iterator()
  {
    return new ByteArrayIterator(start);
  }

  @Override public int lastIndexOf(final Object o)
  {
    if (!(o instanceof Byte)) return -1;
    return byte_lastIndexOf((Byte) o);
  }
  
  public int byte_lastIndexOf(final byte b) {
    // Since Byte is final, make this little bit more efficient
    // Could be done even more efficiently by preallocating array of 256 
    // elements
    for (int i=size-1;i>=0; i--) if(arr[start+i]==b) return i;
    return -1;
  }

  @Override public ListIterator<Byte> listIterator()
  {
    return new ByteArrayIterator(start);
  }

  @Override public ListIterator<Byte> listIterator(final int idx)
  {
    checkIdx(idx);
    return new ByteArrayIterator(start+idx);
  }

  /**
   * Not supported
   */
  @Override public boolean remove(final Object o)
  {
    throw new UnsupportedOperationException();
  }

  /**
   * Not supported
   */
  @Override public Byte remove(final int index)
  {
    throw new UnsupportedOperationException();
  }

  /**
   * Not supported
   */
  @Override public boolean removeAll(final Collection<?> coll)
  {
    throw new UnsupportedOperationException();
  }

  /**
   * Not supported
   */
  @Override public boolean retainAll(final Collection<?> coll)
  {
    throw new UnsupportedOperationException();
  }

  @Override public Byte set(final int idx, final Byte b)
  {
    return byte_set(idx, b);
  }

  public byte byte_set (final int idx, final byte b) {
    checkIdx(idx);
    final byte old = arr[start+idx];
    arr[start+idx] = b;
    return old;
  }
  
  public void byte_set (final int idx, final byte[] bytes) {
    byte_set(idx, bytes, 0, bytes.length);
  }
  
  public void byte_set (
      final int idx, 
      final byte[] bytes, 
      final int srcFromIdx,
      final int srcToIdx
  ) {
    checkIdx(idx);
    final int len = srcToIdx-srcFromIdx;
    if (len>size-idx) throw new IllegalArgumentException(
        "bytes array is too big (srcToIdx-srcFromIdx="+len+">size-idx="
        +(size-idx));
    System.arraycopy(bytes, srcFromIdx, arr, idx, srcToIdx-srcFromIdx);
  }

  @Override public int size()
  {
    return size;
  }

  @Override public List<Byte> subList(final int fromIdx, final int toIdx)
  {
    if (fromIdx>toIdx) throw new IllegalArgumentException(
        "fromIdx="+fromIdx+">toIdx="+toIdx); 
    checkIdx(fromIdx);
    checkIdx(toIdx-1);
    return new FixedByteArray(arr, start+fromIdx, start+toIdx);
  }

  @Override public Object[] toArray()
  {
    final Byte [] res = new Byte[size];
    for (int i=0; i<size; i++) res[i] = arr[start+i];
    return res;
  }

  @SuppressWarnings("unchecked") @Override 
    public <T> T[] toArray(final T[] res)
  {
    for (int i=0; i<size; i++) res[i] = (T) ((Byte) arr[start+i]);
    return null;
  }

  protected void checkIdx(final int idx) throws IndexOutOfBoundsException 
  {
    if (idx<0 || idx>=size) 
        throw new IndexOutOfBoundsException("idx="+idx+", size="+size);
  }

  @Override public String toString()
  {
    final StringBuilder res = new StringBuilder();
    res.append("[");
    boolean first = true;
    for (int i=0; i<size; i++) {
      if (!first) {
        res.append(", ");
      } else first = false;
      res.append(String.valueOf(arr[start+i]));
    }
    res.append("]");
    return res.toString();
  }
  
  private class FixedByteArrayInputStream extends InputStream {

    int idx = 0;
    int markIdx = -1;
    

    @Override public int available() throws IOException
    {
      return size-idx;
    }

    @Override public synchronized void mark(final int readlimit)
    {
      markIdx = idx;
    }

    @Override public boolean markSupported()
    {
      return true;
    }

    @Override public int read(final byte[] b, final int off, final int len) 
      throws IOException
    {
      if (idx >= size) return -1;
      final int toRead = len < size-idx ? len : size-idx; 
      System.arraycopy(arr, idx, b, off, toRead);
      idx+=toRead;
      return toRead;
    }

    @Override public int read(final byte[] b) throws IOException
    {
      return read(b, 0, b.length);
    }

    @Override public synchronized void reset() throws IOException
    {
      if (markIdx<0) throw new IOException("mark() was never called");
      idx = markIdx;
    }

    @Override public long skip(final long n) throws IOException
    {
      idx += n;
      return idx;
    }

    @Override public int read() throws IOException
    {
      if (idx >= size) return -1;
      return arr[start+(idx++)];
    }

    @Override public void close() throws IOException {}
  }

  /**
   * Treat this array as input stream
   */
  public InputStream asInputStream() {
    return new FixedByteArrayInputStream();
  }
}
