/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.collections.bytearray;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;

import junit.framework.Assert;

import org.builtu.collections.bytearray.FixedByteArray;
import org.builtu.exceptions.FatalError;
import org.junit.Test;


public class TestFixedByteArraySpecific extends TestFixedByteArray {

  @Test public void test_setb() {
    final byte [] arr = new byte[1];
    final FixedByteArray tst = new FixedByteArray(arr);
    Assert.assertEquals(0, tst.byte_set(0, (byte)100));
    Assert.assertEquals(100, arr[0]);
  }

  @Test public void test_equals_different_offsets() {
    final byte [] arr1 = new byte[]{0,1,2},
                  arr2 = new byte[]{1,2};
    final FixedByteArray a1 = new FixedByteArray(arr1, 1, 3);
    final FixedByteArray a2 = new FixedByteArray(arr2);
    Assert.assertEquals(a1, a2);
  }

  
  @Test public void test_byte_indexof() {
    final byte [] arr1 = new byte[]{0,1,2};
    final FixedByteArray a1 = new FixedByteArray(arr1);
    Assert.assertEquals(0, a1.byte_indexOf((byte)0));
    Assert.assertEquals(1, a1.byte_indexOf((byte)1));
    Assert.assertEquals(2, a1.byte_indexOf((byte)2));
    Assert.assertEquals(-1, a1.byte_indexOf((byte)3));
  }
  
  @Test public void test_byte_lastIndexOf() {
    final byte [] arr1 = new byte[]{0,1,2};
    final FixedByteArray a1 = new FixedByteArray(arr1);
    Assert.assertEquals(0, a1.byte_lastIndexOf((byte)0));
    Assert.assertEquals(1, a1.byte_lastIndexOf((byte)1));
    Assert.assertEquals(2, a1.byte_lastIndexOf((byte)2));
    Assert.assertEquals(-1, a1.byte_lastIndexOf((byte)3));
  }
  
  
  @Test public void test_setbytes() {
    final byte [] arr = new byte[2], bytes = new byte[] {1, 2};
    final FixedByteArray tst = new FixedByteArray(arr);
    tst.byte_set(0, bytes);
    this.assertArraysEquals(bytes, arr);
  }

  @Test public void test_setbytes2() {
    final byte [] arr = new byte[2], bytes = new byte[] {0, 1, 2, 3};
    final FixedByteArray tst = new FixedByteArray(arr);
    tst.byte_set(0, bytes, 1, 3);
    Assert.assertEquals(1, arr[0]);
    Assert.assertEquals(2, arr[1]);
  }

  @Test public void test_containsb() {
    final byte [] arr = new byte[] {0, 1, 2, 3};
    final FixedByteArray tst = new FixedByteArray(arr);
    Assert.assertTrue(tst.byte_contains((byte)0));
    Assert.assertTrue(tst.byte_contains((byte)1));
    Assert.assertTrue(tst.byte_contains((byte)2));
    Assert.assertTrue(tst.byte_contains((byte)3));
    Assert.assertFalse(tst.byte_contains((byte)4));
  }
  
  @Test public void test_byte_get() {
    final byte [] arr = new byte[] {0, 1, 2, 3};
    final FixedByteArray tst = new FixedByteArray(arr);
    Assert.assertEquals(0, tst.byte_get(0));
    Assert.assertEquals(1, tst.byte_get(1));
    Assert.assertEquals(2, tst.byte_get(2));
    Assert.assertEquals(3, tst.byte_get(3));
    try {
      tst.byte_get(4);
      fail();
    } catch (final IndexOutOfBoundsException e) {}
  }

  @Override protected ListOperation<Byte> makeModifier(final List<Byte> ref)
  {
    if (rnd(2)==0) return super.makeModifier(ref);
    final int idx = -2 + rnd(ref.size()+3);
    final Byte e = makeElement();
    final Byte existingElement = ref.size()==0 ? 
                   makeElement() : ref.get(rnd(ref.size()));
    
    
    
    final int opCode = rnd(5);
    switch (opCode) {
    case 0: return new ListOperation<Byte>(){
      @Override public Byte modifyList(final List<Byte> list)
      {
        if (list instanceof FixedByteArray)
          return ((FixedByteArray) list).byte_set(idx, e);
        else
          return list.set(idx, e);
      }}; 
    case 1: return new ListOperation<Byte>(){
      @Override public Byte modifyList(final List<Byte> list)
      {
        if (list instanceof FixedByteArray)
          return b2e(((FixedByteArray) list).byte_indexOf(existingElement)>0, e);
        else
          return b2e(list.indexOf(existingElement)>0, e);
      }}; 
    case 2: return new ListOperation<Byte>(){
      @Override public Byte modifyList(final List<Byte> list)
      {
        if (list instanceof FixedByteArray)
          return b2e(((FixedByteArray) list).byte_indexOf(e)>0, e);
        else
          return b2e(list.indexOf(e)>0, e);
      }}; 
    case 3: return new ListOperation<Byte>(){
      @Override public Byte modifyList(final List<Byte> list)
      {
        if (list instanceof FixedByteArray)
          return b2e(((FixedByteArray) list)
              .byte_lastIndexOf(existingElement)>0, e);
        else
          return b2e(list.lastIndexOf(existingElement)>0, e);
      }}; 
    case 4: return new ListOperation<Byte>(){
      @Override public Byte modifyList(final List<Byte> list)
      {
        if (list instanceof FixedByteArray)
          return b2e(((FixedByteArray) list)
               .byte_lastIndexOf(e)>0, e);
        else
          return b2e(list.lastIndexOf(e)>0, e);
      }}; 
      
    default: fail("Bad opcode "+opCode);
    }
    throw new FatalError("Can't reach here!!!");
  }

  @Test public void testAsInputStream() throws IOException {
    final byte [] data = "Hello, world".getBytes();
    final FixedByteArray ba = new FixedByteArray(data);
    final InputStream in = ba.asInputStream();
    final byte [] clone = new byte[data.length];
    Assert.assertEquals(data.length, in.read(clone));
    Assert.assertEquals(-1, in.read(clone));
    this.assertArraysEquals(data, clone);
  }
  
  @Test public void testAsInputStreamMark() throws IOException {
    final byte [] chunk1 = "Hello".getBytes(),
                  chunk2 = " world!".getBytes();
    final byte [] data = new byte[chunk1.length+chunk2.length];
    System.arraycopy(chunk1, 0, data, 0, chunk1.length);
    System.arraycopy(chunk2, 0, data, chunk1.length, chunk2.length);
    final FixedByteArray ba = new FixedByteArray(data);
    final InputStream in = ba.asInputStream();
    final byte [] buf = new byte[1000];
    Assert.assertEquals(chunk1.length, in.read(buf, 0, chunk1.length));
    Assert.assertTrue(in.markSupported());
    in.mark(0);
    Assert.assertEquals(chunk2.length, in.read(buf));
    final byte [] chunk2copy = new byte[chunk2.length];
    in.reset();
    Assert.assertEquals(chunk2.length, in.read(chunk2copy));
    Assert.assertEquals(-1, in.read(buf));
    this.assertArraysEquals(chunk2, chunk2copy);
  }
}
