/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.collections.testlist;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

import org.builtu.exceptions.FatalError;
import org.junit.Test;


public abstract class AbstractTestDynamicList<E> extends AbstractTestFixedList<E> {
  
  @Test public void testClear() {
    final List<E> ref = makeRef(); 
    testClear(ref);
    testClear(makeTestee(ref));
  }

  private void testClear(final List<E> list)
  {
    list.clear();
    assertEquals("List cannot be cleared", 0, list.size());

    try {
      list.get(-1);
      fail();
    } catch (final IndexOutOfBoundsException e) {}
    try {
      list.get(0);
      fail();
    } catch (final IndexOutOfBoundsException e) {}
    try {
      list.get(1);
      fail();
    } catch (final IndexOutOfBoundsException e) {}
  }

  @Test public void testIteratorRemoveNext() {
    final List<E> ref = makeRef(); 
    testIteratorRemoveNext(ref);
    testIteratorRemoveNext(makeTestee(ref));
  }

  private void testIteratorRemoveNext(final List<E> list)
  {
    list.clear();
    final E e1 = makeElement();
    final E e2 = makeAnother(e1);
    list.add(e1);
    list.add(e2);
    final Iterator<E> it = list.iterator();
    assertEquals(e1, it.next());
    it.remove();
    assertEquals(1, list.size());
    assertEquals(e2, list.get(0));
  }
  
  @Test public void testIteratorRemovePrevious() {
    final List<E> ref = makeRef(); 
    testIteratorRemovePrevious(ref);
    testIteratorRemovePrevious(makeTestee(ref));
  }
  
  private void testIteratorRemovePrevious(final List<E> list)
  {
    list.clear();
    final E e1 = makeElement();
    final E e2 = makeAnother(e1);
    list.add(e1);
    list.add(e2);
    final ListIterator<E> it = list.listIterator();
    assertEquals(e1, it.next());
    assertEquals(e2, it.next());
    assertEquals(e2, it.previous());
    assertEquals(e1, it.previous());
    it.remove();
    assertEquals(1, list.size());
    assertEquals(e2, list.get(0));
  }  
  
  @Test public void testIteratorNextAfterRemove() {
    final List<E> ref = makeRef(); 
    testIteratorNextAfterRemove(ref);
    testIteratorNextAfterRemove(makeTestee(ref));
  }
 
  private void testIteratorNextAfterRemove(final List<E> list)
  {
    list.clear();
    final E e1 = makeElement();
    final E e2 = makeAnother(e1);
    final E e3 = makeAnother(e1, e2);
    list.add(e1);
    list.add(e2);
    list.add(e3);
    final ListIterator<E> it = list.listIterator();
    assertEquals(e1, it.next());
    it.remove();
    assertEquals(e2, it.next());
  }

  @Test public void testIteratorPreviousAfterRemove() {
    final List<E> ref = makeRef(); 
    testIteratorPreviousAfterRemove(ref);
    testIteratorPreviousAfterRemove(makeTestee(ref));
  }
  
  private void testIteratorPreviousAfterRemove(final List<E> list)
  {
    list.clear();
    final E e1 = makeElement();
    final E e2 = makeAnother(e1);
    final E e3 = makeAnother(e1, e2);
    list.add(e1);
    list.add(e2);
    list.add(e3);
    final ListIterator<E> it = list.listIterator();
    assertEquals(e1, it.next());
    assertEquals(e2, it.next());
    assertEquals(e3, it.next());
    it.remove();
    assertEquals(e2, it.previous());
  }

  @Test public void testOneElement() {
    final List<E> ref = makeRef(); 
    testOneElement(ref);
    testOneElement(makeTestee(ref));
  }

  private void testOneElement(final List<E> list)
  {
    list.clear();
    assertEquals("List cannot be cleared", 0, list.size());
    
    final E e = makeElement();

    list.add(e);
    assertEquals("Cannot add element to a list, size unchanged", 
        1, list.size());
    assertSame(e, list.get(0));
    list.remove(e);
    assertEquals(0, list.size());
    list.add(e);
    assertEquals(1, list.size());
    list.remove(0);
    assertEquals(0, list.size());
    list.add(e);

    try {
      list.get(-1);
      fail();
    } catch (final IndexOutOfBoundsException ex) {}
    
    try {
      list.get(1);
      fail();
    } catch (final IndexOutOfBoundsException ex) {}
    
    list.clear();
    assertEquals("List cannot be cleared", 0, list.size());
  }
  
  @Test public void testTwoElements() {
    final List<E> ref = makeRef(); 
    testTwoElements(ref);
    testTwoElements(makeTestee(ref));
  }

  private void testTwoElements(final List<E> list)
  {
    list.clear();
    assertEquals("List cannot be cleared", 0, list.size());
    final E e1 = makeElement();
    final E e2 = makeAnother(e1);

    list.add(e1);
    assertEquals("Cannot add 1st element to a list, size unchanged", 
        1, list.size());

    list.add(e2);
    assertEquals("Cannot add 2nd element to a list, size unchanged", 
        2, list.size());
    
    assertSame(e1, list.get(0));
    assertSame(e2, list.get(1));
    list.remove(e1);
    assertEquals(1, list.size());
    list.remove(e2);
    assertEquals(0, list.size());
    list.add(e1);
    list.add(e2);

    try {
      list.get(-1);
      fail();
    } catch (final IndexOutOfBoundsException e) {}
    
    try {
      list.get(2);
      fail();
    } catch (final IndexOutOfBoundsException e) {}
  }
  
  @Test public void testRangeExclusive() {
    final List<E> ref = makeRef();
    final List<E> tst = makeTestee(ref);
    testRangeExclusive(ref);
    testRangeExclusive(tst);
  }

  @Test public void testAddIndex() {
    final List<E> ref = makeRef();
    final List<E> tst = makeTestee(ref);
    testAddIndex(ref);
    testAddIndex(tst);
  }
  
  private void testAddIndex(final List<E> list) {
    list.add(list.size(), makeElement());
  }
  
  @Test public void testAddAllIndex() {
    final List<E> ref = makeRef();
    final List<E> tst = makeTestee(ref);
    testAddAllIndex(ref);
    testAddAllIndex(tst);
  }

  private void testAddAllIndex(final List<E> list)
  {
    final List<E> coll = new ArrayList<E>();
    coll.add(makeElement());
    list.addAll(list.size(), coll);
  }

  private void testRangeExclusive(final List<E> list)
  {
    list.clear();
    // Following call may treated as wrong or as valid
    // depending on the implementation
    try {
      assertEquals(0, list.subList(0, 0).size());
    } catch (final IndexOutOfBoundsException e) {}
    
    list.add(makeElement());
    assertEquals(1, list.subList(0, 1).size());
  }
  
  
  @Override protected ListOperation<E> makeModifier(final List<E> ref) {
    final int fixedOrDynamic = rnd(2);
    if (fixedOrDynamic == 0) return super.makeModifier(ref);

    final int idx = -2 + rnd(ref.size()+3);
    final E e = makeElement();
    final E existingElement = ref.size()==0 ? 
          makeElement() : ref.get(rnd(ref.size()));    
    final List<E> coll         = randomColl(ref),
                  existingColl = randomCollectionOfExistingElements(ref);
    final int skip = rnd(ref.size()), skip2 = rnd(ref.size()/2);
    
    
    final int opCode = rnd(18);
    switch(opCode) {
    case 0:
    case 1: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        return b2e(list.add(e), e);
      }};
    case 2: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        return b2e(list.add(existingElement), existingElement);
      }};
    case 3: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        return b2e(list.addAll(coll), e);
      }};
    case 4: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        return b2e(list.addAll(existingColl), e);
      }};
    case 5: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        return b2e(list.addAll(idx, coll), e);
      }};
    case 6: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        return b2e(list.addAll(idx, existingColl), e);
      }};
    case 7: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        list.iterator().remove();
        return list.iterator().next();
      }};
    case 8: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        final Iterator<E> it = list.iterator();
        for (int i=1; i<skip; i++) it.next();
        it.remove();
        return it.next();
      }};
    case 9: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        final ListIterator<E> it = list.listIterator();
        for (int i=1; i<=skip; i++) it.next();
        for (int i=1; i<=skip2; i++) it.previous();
        it.remove();
        return it.next();
      }};
    case 10: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        list.clear();
        return null;
      }};
    case 11: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        return list.remove(idx);
      }};
    case 12: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        return b2e(list.remove(e), e);
      }};
    case 13: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        return b2e(list.remove(existingElement), existingElement);
      }};
    case 14: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        return b2e(list.removeAll(coll), e);
      }};
    case 15: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        return b2e(list.removeAll(existingColl), e);
      }};
    case 16: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        return b2e(list.retainAll(coll), e);
      }};
    case 17: return new ListOperation<E>(){
      @Override public E modifyList(final List<E> list)
      {
        return b2e(list.retainAll(existingColl), e);
      }};
    default: fail("Bad opcode "+opCode);
    }
    throw new FatalError("Can't reach here!!!");
  }

}
