/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.testing;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Random;
import java.util.Set;

import org.junit.After;
import org.junit.Assert;
import org.junit.Before;

/**
 * This basic testcase supports several useful features:
 * <ul>
 * <li>random testing</li>
 * <li>temporary files and directories</li>
 * <li>basic serialization test {@link BasicTestCase#assertSerializable}</li>
 * </ul>
 * 
 * @author borisman
 * 
 */
public class BasicTestCase {
	private final int STRING_MAXLENGTH = 100;

	// START random testing support ====================================================================================
	private long seed;
	protected final Random random = new Random();

	@Before	public void initSeed() {
		setSeed(System.currentTimeMillis());
	}

	protected void setSeed(final long seed) {
		this.seed = seed;
		System.out.println(seedInfo());
		this.random.setSeed(seed);
	}

	protected void reseed() {
		setSeed(rnd());
	}

	/**
	 * @return string, containing random characters from whole char range (including bad, unmapped, unprintable chars).
	 *         The length of the string is in range [0: STRING_MAXLENGTH]
	 * @see BasicTestCase#rndString(int, int)
	 */
	protected String rndString() {
		return rndString(0, STRING_MAXLENGTH);
	}

	/**
	 * @return string, containing random US-ASCII characters. The length of the string is in 
	 * range [0: STRING_MAXLENGTH].
	 * @see BasicTestCase#rndAsciiString(int, int)
	 */
	protected String rndAsciiString() {
		return rndAsciiString(0, STRING_MAXLENGTH);
	}

	/**
	 * @return string, containing random printable characters. The length of the string is in
	 * range [0: STRING_MAXLENGTH]
	 * @see BasicTestCase#rndAsciiString(int, int)
	 */
	protected String rndPrintableString() {
		return rndPrintableString(0, STRING_MAXLENGTH);
	}

	/**
	 * @return string, containing random alphanumeric characters. The length of the string is in
	 * range [0: STRING_MAXLENGTH]
	 * @see BasicTestCase#rndAlnumString(int, int)
	 */
	protected String rndAlnumString() {
		return rndAlnumString(0, STRING_MAXLENGTH);
	}

	/**
	 * @return string, containing random alpha characters. The length of the string is in
	 * range [0: STRING_MAXLENGTH]
	 * @see BasicTestCase#rndAlphaString(int, int)
	 */
	protected String rndAlphaString() {
		return rndAlphaString(0, STRING_MAXLENGTH);
	}

	/**
	 * @return java identifier consisting of random characters (first char is alpha, rest is alphanumeric). 
	 * The length of the string is in range [1: STRING_MAXLENGTH]
	 * @see BasicTestCase#rndIdentifier(int, int)
	 */
	protected String rndIdentifier() {
		return rndIdentifier(1, STRING_MAXLENGTH);
	}

	/**
	 * Internal interface for {@link BasicTestCase#rndString} function
	 * @author borisman
	 *
	 */
	protected interface RndCharGenerator {
		char rndChar();
	}

	/**
	 * Basic {@link BasicTestCase#RndCharGenerator} implementation that selects random character from provided string
	 * @author borisman
	 *
	 */
	protected final class RndCharSelectGenerator implements RndCharGenerator {
		private final String charactersSet;

		public RndCharSelectGenerator(final String charactersSet) {
			this.charactersSet = charactersSet;
		}

		public char rndChar() {
			final int index = random.nextInt(charactersSet.length());
			return charactersSet.charAt(index);
		}
	}

	private static final String RND_PRINTABLE_CHARS = "`1234567890-=\\qwertyuiop[]asdfghjkl;'zxcvbnm,./~!@#$%^&*()_+|"+
			                                          "QWERTYUIOP{}ASDFGHJKL:\"ZXCVBNM<>?";
	private static final String RND_ALPHA_CHARS = "qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM";
	private static final String RND_ALPHANUM_CHARS = RND_ALPHA_CHARS+"1234567890";

	protected final RndCharGenerator printableCharGenerator = new RndCharSelectGenerator(RND_PRINTABLE_CHARS);
	protected final RndCharGenerator alphaCharGenerator = new RndCharSelectGenerator(RND_ALPHA_CHARS);
	protected final RndCharGenerator alnumCharGenerator = new RndCharSelectGenerator(RND_ALPHANUM_CHARS);

	protected final RndCharGenerator asciiCharGenerator = new RndCharGenerator() {
		public char rndChar() {
			return (char) (-127 + random.nextInt(129));
		}
	};

	protected final RndCharGenerator stdCharGenerator = new RndCharGenerator() {
		public char rndChar() {
			return (char) random.nextInt();
		}
	};

	
	/**
	 * @return string, containing random printable characters. The length of the string is in range [minlen:maxlen]
	 */
	protected String rndPrintableString(final int minlen, final int maxlen) {
		return rndString(minlen, maxlen, printableCharGenerator);
	}

	/**
	 * @return java identifier consisting of random characters (first char is alpha, rest is alphanumeric). 
	 * The length of the string is in range [minlen:maxlen]
	 * @see BasicTestCase#rndIdentifier(int, int)
	 */
	protected String rndIdentifier(final int minlen, final int maxlen) {
		if (minlen > maxlen)
			throw new IllegalArgumentException("minlen>maxlen");
		final int dev = maxlen - minlen;
		final int len = (dev <= 0) ? minlen : minlen + random.nextInt(dev);
		if (len <= 0) {
			return "";
		} else {
			return alphaCharGenerator.rndChar()	+ rndAlnumString(len - 1, len - 1);
		}
	}

	/**
	 * @return string, containing random alpha characters. The length of the string is in range [minlen:maxlen]
	 */
	protected String rndAlphaString(final int minlen, final int maxlen) {
		return rndString(minlen, maxlen, alphaCharGenerator);
	}

	/**
	 * @return string, containing random alphanumeric characters. The length of the string is in range [minlen:maxlen]
	 */
	protected String rndAlnumString(final int minlen, final int maxlen) {
		return rndString(minlen, maxlen, alnumCharGenerator);
	}

	/**
	 * @return string, containing random US-ASCII characters. The length of the string is in range [minlen: maxlen].
	 */
	protected String rndAsciiString(final int minlen, final int maxlen) {
		return rndString(minlen, maxlen, asciiCharGenerator);
	}

	/**
	 * @return string, containing random characters from whole char range (including bad, unmapped, unprintable chars).
	 *         The length of the string is in range [minlen:maxlen]
	 */
	protected String rndString(final int minlen, final int maxlen) {
		return rndString(minlen, maxlen, stdCharGenerator);
	}

	/**
	 * @return random string with length in range [minlen:maxlen] consisting of characters obtained from generator
	 */
	protected String rndString(final int minlen, final int maxlen, final RndCharGenerator generator) {
		if (minlen > maxlen) throw new IllegalArgumentException("minlen>maxlen");
		final int dev = maxlen - minlen;
		final int len = (dev <= 0) ? minlen : minlen + random.nextInt(dev);
		final StringBuilder result = new StringBuilder();
		for (int i = 0; i < len; i++) {
			final char c = generator.rndChar();
			result.append(c);
		}
		return result.toString();
	}

	/**
	 * @return random integer from [0, max) if max&gt;=0 or [-max, 0) if max&lt;0
	 */
	protected int rnd(final int max) {
		return random.nextInt(max < 0 ? -max : max == 0 ? 1 : max); // java.lang.IllegalArgumentException:
																	// max must
																	// be
																	// positive
	}

	/**
	 * @return random integer in range [{@link Integer#MIN_VALUE}:{@link Integer#MAX_VALUE}]
	 */
	protected int rnd() {
		return random.nextInt();
	}

	/**
	 * @return random element from list
	 * @throws IllegalArgumentException when list is empty 
	 */
	protected <E> E rndChoice(final List<E> list) {
		final int N = list.size();
		if (N == 0) throw new IllegalArgumentException("Empty list");
		final int idx = rnd(N);
		return list.get(idx);
	}

	/**
	 * @return random element from collection
	 * @throws IllegalArgumentException when coll is empty 
	 */
	protected <E> E rndChoice(final Collection<E> coll) {
		final int N = coll.size();
		if (N == 0) throw new IllegalArgumentException("Empty collection");
		
		final Iterator<E> it = coll.iterator();
		int left = N;
		for (;;) {
			final E current = it.next();
			
			final int roll = this.rnd(left);
			// random chance is = 1/left
			final boolean currentElementSelected = roll == 0;
	
			if (currentElementSelected) return current;
		}
	}
	
	/**
	 * @return random element from array
	 * @throws IllegalArgumentException when array is empty 
	 */	
	protected <E> E rndChoice(final E[] array) {
		if (array.length == 0) throw new IllegalArgumentException("Empty array");
		
		final int idx = rnd(array.length);
		return array[idx];
	}

	/**
	 * @return random character from whole char range (including unmapped and unprintable chars)
	 */
	protected char rndChar() {
		return (char) (random.nextInt());
	}
	
	/**
	 * @return random alpha character
	 */
	protected char rndAlphaChar() {
		return alphaCharGenerator.rndChar();
	}

	protected float rndFloat() {
		return random.nextFloat();
	}

	protected double rndDouble() {
		return random.nextDouble();
	}

	protected long rndLong() {
		return random.nextLong();
	}

	/**
	 * Fill strs with random strings
	 * @see BasicTestCase#rndString()
	 */
	protected void rndStrings(final String[] strs) {
		for (int i = 0; i < strs.length; i++) strs[i] = rndString();
	}

	/**
	 * Fill strs with random alphanumeric strings
	 * @see BasicTestCase#rndAlnumString()
	 */
	protected void rndAlnumStrings(final String[] strs) {
		for (int i = 0; i < strs.length; i++)
			strs[i] = rndAlnumString();
	}

	/**
	 * Fill strs with uniuque (non-equal) alphanumeric strings
	 * @see BasicTestCase#rndAlnumString()
	 */
	
	protected void rndUniqueAlnumStrings(final String[] strs) {
		final List<String> uniqueStrings = new ArrayList<String>();
		for (int i = 0; i < strs.length; i++) {
			String candidate = rndAlnumString();
			while (uniqueStrings.indexOf(candidate) >= 0)
				candidate = rndAlnumString();
			uniqueStrings.add(candidate);
		}
		uniqueStrings.toArray(strs);
	}

	/**
	 * Fill strs with random alpha strings
	 * @see BasicTestCase#rndAlphaString()
	 */
	protected void rndAlphaStrings(final String[] strs) {
		for (int i = 0; i < strs.length; i++)
			strs[i] = rndAlphaString();
	}

	/**
	 * Fill strs with random US-ASCII strings
	 * @see BasicTestCase#rndAsciiString()
	 */
	protected void rndAsciiStrings(final String[] strs) {
		for (int i = 0; i < strs.length; i++)
			strs[i] = rndAsciiString();
	}

	/**
	 * Fill strs with random java identifiers strings
	 * @see BasicTestCase#rndIdentifiers()
	 */
	protected void rndIdentifiers(final String[] strs) {
		for (int i = 0; i < strs.length; i++)
			strs[i] = rndIdentifier();
	}


	protected void assertEquals(final String message, final Object expected, final Object actual) {
		Assert.assertEquals(message + seedInfo(), expected, actual);
	}

	protected void assertEquals(final float expected, final float actual, final float delta) {
		Assert.assertEquals("", expected, actual, delta);
	}

	protected void assertEquals(final double expected, final double actual,	final double delta) {
		Assert.assertEquals("", expected, actual, delta);
	}

	protected void assertEquals(final String message, final float expected,	final float actual, final float delta) {
		Assert.assertEquals(message + seedInfo(), expected, actual, delta);
	}

	protected void assertEquals(final Object expected, final Object actual) {
		assertEquals("", expected, actual);
	}

	protected void assertSame(final Object expected, final Object actual) {
		assertSame("", expected, actual);
	}

	protected void assertSame(final String message, final Object expected,
			final Object actual) {
		Assert.assertEquals(message + seedInfo(), expected, actual);
	}

	protected void assertTrue(final String message, final boolean condition) {
		Assert.assertTrue(message + seedInfo(), condition);
	}

	protected void assertNotNull(final String message, final Object condition) {
		Assert.assertNotNull(message + seedInfo(), condition);
	}

	protected void assertNull(final String message, final Object condition) {
		Assert.assertNull(message + seedInfo(), condition);
	}

	protected void assertTrue(final boolean condition) {
		assertTrue("", condition);
	}

	protected void assertFalse(final String message, final boolean condition) {
		Assert.assertFalse(message + seedInfo(), condition);
	}

	protected void assertFalse(final boolean condition) {
		assertFalse("", condition);
	}

	protected void fail() {
		Assert.fail(seedInfo());
	}

	protected void fail(final String message) {
		Assert.fail(message + seedInfo());
	}

	protected void fail(final Exception e) {
		throw new RuntimeException("Failed " + seedInfo(), e);
	}

	protected void assertIteratorsElementsEquals(final Iterator<?> it1,	final Iterator<?> it2) {
		int pos = 0;
		while (it1.hasNext()) {
			assertTrue("it2 has no more elements", it2.hasNext());
			final Object e1 = it1.next();
			final Object e2 = it2.next();
			assertEquals("mismatch at position " + pos, e1, e2);
			pos++;
		}
		Assert.assertFalse("it2 has more elements than it1", it2.hasNext());
		try {
			it1.next();
			Assert.fail("it1 does not throw NoSuchElementException");
		} catch (final NoSuchElementException e) {
			e.printStackTrace();
		}
		try {
			it2.next();
			Assert.fail("it2 does not throw NoSuchElementException");
		} catch (final NoSuchElementException e) {
			e.printStackTrace();
		}
	}

	protected void assertIterablesElementsEquals(final Iterable<?> it1,	final Iterable<?> it2) {
		assertIteratorsElementsEquals(it1.iterator(), it2.iterator());
	}

	protected void assertListsEquals(final List<?> it1, final List<?> it2) {
		assertIteratorsElementsEquals(it1.iterator(), it2.iterator());
	}

	protected <E> void assertArraysEquals(final List<E> expected, final List<E> actual) {
		assertEquals("expected.size()!=actual.size()", expected.size(), actual.size());
		for (int i = 0; i < expected.size(); i++) {
			assertEquals("expected[" + i + "]!=actual[" + i + "]", expected.get(i), actual.get(i));
		}
	}

	protected <E> void assertEquals(final Iterator<E> expected, final Iterator<E> actual) {
		assertEquals(expected.hasNext(), actual.hasNext());
		while (expected.hasNext() || actual.hasNext()) {
			assertEquals("both iterators should contains the same elements", expected.next(), actual.next());
		}
		assertEquals("both iterators should contains the same quantity of elements",
				      expected.hasNext(), actual.hasNext());
	}

	protected <E> void assertArraysEquals(final E[] expected, final E[] actual) {
		assertEquals("expected.length!=actual.length", expected.length,	actual.length);
		for (int i = 0; i < expected.length; i++) {
			assertEquals("expected[" + i + "]!=actual[" + i + "]", expected[i], actual[i]);
		}
	}

	protected void assertArraysEquals(final byte[] expected, final byte[] actual) {
		assertEquals("expected.length!=actual.length", expected.length, actual.length);
		for (int i = 0; i < expected.length; i++) {
			assertEquals("expected[" + i + "]!=actual[" + i + "]", expected[i], actual[i]);
		}
	}
	protected void assertArraysEquals(final int[] expected, final int[] actual) {
		assertEquals("expected.length!=actual.length", expected.length,	actual.length);
		for (int i = 0; i < expected.length; i++) {
			assertEquals("expected[" + i + "]!=actual[" + i + "]", expected[i], actual[i]);
		}
	}
	

	/**
	 * @return current seed. Useful for rerunning testcase with it by placing {@link #setSeed(long)} at the start of the
	 *                 testcase.
	 */
	protected String seedInfo() {
		return " (seed=" + seed + "L)";
	}

	// END   random testing support ====================================================================================

	/**
	 * Serialize object into memory and deserialize it back. Useful for
	 * accerting basic serializability (that object does not indirect link to
	 * unserializable object).
	 * 
	 * @return deserialzieed object
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public static Object assertSerializable(final Object o) throws IOException,
			ClassNotFoundException {
		final ByteArrayOutputStream outStream = new ByteArrayOutputStream();
		final ObjectOutputStream out = new ObjectOutputStream(outStream);
		out.writeObject(o);
		final byte[] data = outStream.toByteArray();
		final ByteArrayInputStream inStream = new ByteArrayInputStream(data);
		final ObjectInputStream in = new ObjectInputStream(inStream);
		return in.readObject();
	}

	// START temp files support ========================================================================================

	private File TMP;
	private static final List<File> tempPaths = new ArrayList<File>();

	protected File createTempDir() {
		if (TMP == null)
			TMP = new File(System.getProperty("java.io.tmpdir"));
		File res = new File(TMP, this.getClass().getSimpleName() + rnd());
		while (res.exists() || !res.mkdir()) {
			res = new File(TMP, this.getClass().getSimpleName() + rnd());
		}
		res.deleteOnExit();
		synchronized (tempPaths) {
			tempPaths.add(res);
		}
		return res;
	}

	protected File createTempFile() throws IOException {
		final File res = File.createTempFile(this.getClass().getSimpleName(), ".tmp");
		res.deleteOnExit();
		synchronized (tempPaths) {
			tempPaths.add(res);
		}
		return res;
	}

	@After
	public void deleteTempPaths() {
		synchronized (tempPaths) {
			for (final File f : tempPaths)
				deleteTempPath(f);
		}
	}

	private void deleteTempPath(final File f) {
		if (f.isFile()) {
			f.delete();
		} else if (f.isDirectory()) {
			for (final File sub : f.listFiles())
				deleteTempPath(sub);
			f.delete();
		}
	}

    public static<E> boolean collectionContainsExactly(final Iterable<E> list, final Set<E> expected) {
    	final List<E> l = new ArrayList<E>();
    	for (final E e : list) l.add(e);
    	l.removeAll(expected);
    	return l.size()==0;
    }

    public static<E> boolean collectionContainsExactly(final E[] list, final E ... expected) {
    	final Set<E> set = new HashSet<E>();
    	for (final E e : expected) set.add(e);
    	return collectionContainsExactly(list, set);
    }

    public static<E> boolean collectionContainsExactly(final Iterable<E> list, final E ... expected) {
    	final Set<E> set = new HashSet<E>();
    	for (final E e : expected) set.add(e);
    	return collectionContainsExactly(list, set);
    }
    
    public static<E> boolean collectionContainsExactly(final E[] list, final Set<E> expected) {
    	final List<E> l = new ArrayList<E>();
    	for (final E e : list) l.add(e);
    	l.removeAll(expected);
    	return l.size()==0;
    }
	
	// END   temp files support ========================================================================================

	
	// QUESTIONABLE USABILITY
	/*
	public static <E> boolean arrayContainsExactly(final Iterable<E> list, final Set<E> expected) {
		final List<E> l = new ArrayList<E>();
		for (final E e : list)
			l.add(e);
		l.removeAll(expected);
		return l.size() == 0;
	}
	public static <E> boolean arrayContainsExactly(final E[] list,
			final E... expected) {
		final Set<E> set = new HashSet<E>();
		for (final E e : expected)
			set.add(e);
		return arrayContainsExactly(list, set);
	}

	public static <E> boolean arrayContainsExactly(final Iterable<E> list,
			final E... expected) {
		final Set<E> set = new HashSet<E>();
		for (final E e : expected)
			set.add(e);
		return arrayContainsExactly(list, set);
	}

	public static <E> boolean arrayContainsExactly(final E[] list,
			final Set<E> expected) {
		final List<E> l = new ArrayList<E>();
		for (final E e : list)
			l.add(e);
		l.removeAll(expected);
		return l.size() == 0;
	}
	protected String rndLang() {
		return rndAlphaString(2, 2).toLowerCase();
	}
	protected String rndLang() {
		return rndAlphaString(2, 2).toLowerCase();
	}

	protected final RndCharGenerator noZZZCharGenerator = new RndCharSelectGenerator(
			"1234567890qwertyuiopasdfghjklxcvbnmQWERTYUIOPASDFGHJKLXCVBNM");

	protected String rndNoZZZString(final int minlen, final int maxlen) {
		return rndString(minlen, maxlen, noZZZCharGenerator);
	}




	*/
}
