/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.util.commutil;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.builtu.exceptions.FatalError;
import org.builtu.exceptions.StringFormatException;


public class StringUtil {
    /**
     * Runs {@link Util#split(String, char, int)} with unlimited number of
     * splits (maxSplits=0 means unlimited splitting)
     *
     */
    public static String[] split(final String str, final char c) {
        return split(str, c, 0);
    }

    public static String[] rsplit(final String str, final char c) {
        return rsplit(str, c, 0);
    }

    /**
     * Runs {@link Util#split(String, String, int)} with unlimited number of
     * splits (maxSplits=0 means unlimited splitting)
     *
     */
    public static String[] split(final String str, final String delimeter) {
        return split(str, delimeter, 0);
    }

    public static String[] rsplit(final String str, final String delimeter) {
        return rsplit(str, delimeter, 0);
    }
    
    /**
     * Same as {@link Util#split(String, String, int)}, but optimized for
     * characters
     *
     * @param str
     *            String to split
     * @param c
     *            Split by this character
     * @param maxSplits
     *            maximum number of splits
     */
    public static String[] split(final String str, final char c,
            final int maxSplits) {
        final List<String> splitted = split_list(str, c, maxSplits);
        final String[] res = new String[splitted.size()];
        splitted.toArray(res);
        return res;
    }

    public static String[] rsplit(final String str, final char c,
            final int maxSplits) {
        final List<String> splitted = rsplit_list(str, c, maxSplits);
        final String[] res = new String[splitted.size()];
        splitted.toArray(res);
        return res;
    }
    /**
     * Splits string into separate parts RIGHT way. Due to absence String#join
     * function in Java, the {@link String#split(String)} implementation is
     * broken in following way: <code>"a,b,".split(",")</code> will return
     * <code>{"a", "b"}</code>, not <code>{"a", "b", ""}</code> Now if you join
     * resulting array by <code>","</code> character, you will get
     * <code>"a,b"</code> string, not original <code>"a,b,"</code> string.
     *
     * This function provides right behavior, so when you split
     * <code>Util.split("a,b,cd,,,", ",")</code> you will get
     * <code>{"a", "b", "cd", "", "", ""}</code>
     *
     * This means that <code> str == Util.join(c, Util.split(str, c)) </code>
     * holds true for any character c and string str.
     *
     * maxSplits is used for limiting number of splits, so
     * <code>Util.split("a,b,cd,,,", ",", 1)</code> will give you
     * <code>{"a", "b,cd,,,"}</code>
     *
     * For joining splitted strings,
     *
     * @see Util#join(char, Collection)
     *
     * @param str
     *            String to split
     * @param delimeter
     *            Split by this token
     * @param maxSplits
     *            maximum number of splits
     */
    public static String[] split(final String str, final String delimeter,
            final int maxSplits) {
        final List<String> splitted = split_list(str, delimeter, maxSplits);
        final String[] res = new String[splitted.size()];
        splitted.toArray(res);
        return res;
    }

    public static String[] rsplit(final String str, final String delimeter,
            final int maxSplits) {
        final List<String> splitted = rsplit_list(str, delimeter, maxSplits);
        final String[] res = new String[splitted.size()];
        splitted.toArray(res);
        return res;
    }

    /**
     * Same as {@link Util#split(String, char)}, but returns List&lt;String&gt;
     *
     * @see Util#split_list(String, String)
     */
    public static List<String> split_list(final String str, final char c) {
        return split_list(str, c, 0);
    }

    public static List<String> rsplit_list(final String str, final char c) {
        return rsplit_list(str, c, 0);
    }
    /**
     * Same as {@link Util#split(String, String)}, but returns
     * List&lt;String&gt;
     *
     * @see Util#split_list(String, char)
     */
    public static List<String> split_list(final String str, final String delimeter) {
        return split_list(str, delimeter, 0);
    }

    public static List<String> rsplit_list(final String str, final String delimeter) {
        return rsplit_list(str, delimeter, 0);
    }
    
    /**
     * Same as {@link Util#split(String, char, int)}, but returns
     * List&lt;String&gt;
     *
     * @see Util#split_list(String, String, int)
     */
    public static List<String> split_list(final String str, final char c, final int maxSplits) {
        assert maxSplits >= 0 : "maxSplits<0:" + maxSplits;

        final List<String> res = new ArrayList<String>(maxSplits==0 ? str.length() : maxSplits);
        int pos, prev = 0, splits = 0;
        pos = str.indexOf(c);
        while (pos >= 0) {
            res.add(str.substring(prev, pos));
            prev = pos + 1;
            splits++;
            if (maxSplits > 0 && splits >= maxSplits) break;
            pos = str.indexOf(c, prev);
        }
        res.add(str.substring(prev));
        return res;
    }

    public static List<String> rsplit_list(final String str, final char c, final int maxSplits) {
        assert maxSplits >= 0 : "maxSplits<0:" + maxSplits;

        final List<String> res = new ArrayList<String>(maxSplits==0 ? str.length() : maxSplits);
        int pos, prev = str.length(), splits = 0;
        pos = str.lastIndexOf(c);
        while (pos >= 0) {
            res.add(0, str.substring(pos+1, prev));
            prev = pos;
            splits++;
            if (maxSplits > 0 && splits >= maxSplits) break;
            if (pos==0) break;
            pos = str.lastIndexOf(c, prev-1);
        }
        res.add(0, str.substring(0, prev));
        return res;
    }
    
    /**
     * Same as {@link Util#split(String, String, int)}, but returns
     * List&lt;String&gt;
     *
     * @see Util#split_list(String, char, int)
     */
    public static List<String> split_list(final String str, final String delimeter, final int maxSplits) {
        assert maxSplits >= 0 : "maxSplits<0:" + maxSplits;
        if (delimeter.length() == 0)
            throw new IllegalArgumentException("Delimeter cannot be null");

        final List<String> res = new ArrayList<String>(maxSplits==0 ? str.length() : maxSplits);
        final int len = delimeter.length();
        int pos, prev = 0, splits = 0;
        pos = str.indexOf(delimeter);
        while (pos >= 0) {
            res.add(str.substring(prev, pos));
            prev = pos + len;
            splits++;
            if (maxSplits > 0 && splits >= maxSplits) break;
            pos = str.indexOf(delimeter, prev);
        }
        res.add(str.substring(prev));
        return res;
    }

    public static List<String> rsplit_list(final String str, final String delimeter, final int maxSplits) {
        assert maxSplits >= 0 : "maxSplits<0:" + maxSplits;
        if (delimeter.length() == 0)
            throw new IllegalArgumentException("Delimeter cannot be null");

        final List<String> res = new ArrayList<String>(maxSplits==0 ? str.length() : maxSplits);
        final int len = delimeter.length();
        int pos, prev = str.length(), splits = 0;
        pos = str.lastIndexOf(delimeter);
        while (pos >= 0) {
            res.add(0, str.substring(pos+len, prev));
            prev = pos;
            splits++;
            if (maxSplits > 0 && splits >= maxSplits) break;
            if (pos==0) break;
            pos = str.lastIndexOf(delimeter, prev-1);
        }
        res.add(0, str.substring(0, prev));
        return res;
    }

    /**
     * Joins collection of strings by specified character, so
     * <code>Util.join("-", {"a", "b", ""})</code> will give you
     * <code>"a-b-"</code>
     *
     * See {@link Util#split(String, char)},
     * {@link Util#split(String, char, int)},
     */
    public static String join(final char c, final Collection<?> components) {
        final StringBuilder res = new StringBuilder(components.size() << 4);
        boolean first = true;
        for (final Object component : components) {
            if (!first) res.append(c);
            res.append(component.toString());
            first = false;
        }
        return res.toString();
    }

    /**
     * Joins collection of strings by specified character, so
     * <code>Util.join('-', {"a", "b", ""})</code> will give you
     * <code>"a-b-"</code>
     *
     * @see Util#join(String, Object[])
     */
    public static String join(final char c, final Object[] components) {
        // try to guess resulting string size, assume average string is 16
        // elements
        final StringBuilder res = new StringBuilder(components.length << 4);
        boolean first = true;
        for (final Object component : components) {
            if (!first) res.append(c);
            res.append(component.toString());
            first = false;
        }
        return res.toString();
    }

    /**
     * Joins collection of objects by specified string, so
     * <code>Util.join("---", {"a", "b", ""})</code> will give you
     * <code>"a---b---"</code>
     *
     * @see Util#join(char, Collection)
     */
    public static String join(final String s, final byte[] components) {
        // try to guess resulting string size, assume average string is 3
        // elements
        final int concats = components.length == 0 ? 0 : components.length - 1;
        final int estimation = (components.length * 3) + concats*s.length();
        assert estimation >= 0;
        final StringBuilder res = new StringBuilder(estimation);
        boolean first = true;
        for (final byte component : components) {
            if (!first)
                res.append(s);
            res.append(String.valueOf(component));
            first = false;
        }
        return res.toString();
    }

    /**
     * Joins collection of integers by specified string, so
     * @see Util#join(char, Collection)
     */
    public static String join(final String s, final int[] components) {
        // try to guess resulting string size, assume average string is 3
        // elements
        final int concats = components.length == 0 ? 0 : components.length - 1;
        final int estimation = (components.length * 3) + concats*s.length();
        assert estimation >= 0;
        final StringBuilder res = new StringBuilder(estimation);
        boolean first = true;
        for (final int component : components) {
            if (!first)
                res.append(s);
            res.append(String.valueOf(component));
            first = false;
        }
        return res.toString();
    }
    
    /**
     * Joins collection of objects by specified string, so
     * <code>Util.join("---", {"a", "b", ""})</code> will give you
     * <code>"a---b---"</code>
     *
     * @see Util#join(char, Collection)
     */
    public static String join(final char c, final int[] components) {
        // try to guess resulting string size, assume average string is 3
        // elements
        final int concats = components.length == 0 ? 0 : components.length - 1;
        final int estimation = (components.length * 3) + concats;
        assert estimation >= 0;
        final StringBuilder res = new StringBuilder(estimation);
        boolean first = true;
        for (final int component : components) {
            if (!first)
                res.append(c);
            res.append(String.valueOf(component));
            first = false;
        }
        return res.toString();
    }
    
    /**
     * Joins collection of strings by specified string, so
     * <code>Util.join("---", {"a", "b", ""})</code> will give you
     * <code>"a---b---"</code>
     *
     * @see Util#join(char, Collection)
     */
    public static String join(final String s, final Object[] components) {
        // try to guess resulting string size, assume average string is 16
        // elements
        final int concats = components.length == 0 ? 0 : components.length - 1;
        final int estimation = (components.length << 4) + concats;
        assert estimation >= 0;
        final StringBuilder res = new StringBuilder(estimation);
        boolean first = true;
        for (Object component : components) {
            if (!first)
                res.append(s);
            res.append(component);
            first = false;
        }
        return res.toString();
    }
    

    public static String join(final String s, final Iterable<?> components) {
        final StringBuilder res = new StringBuilder();
        boolean first = true;
        for (final Object component : components) {
            if (!first) res.append(s);
            res.append(component);
            first = false;
        }
        return res.toString();
    }
    
    /**
     * Escape special characters (CR, LF, Tab, FF) in string according to
     * regular Java (and C) string escaping scheme. This function also encodes
     * quotes (just in case). For example, if you escape <code>ThisCRisLFaTAB\string"'</code>
     * you will get <code>This\ris\na\t\\string\"\'</code>
     *
     *
     * @param str
     *            to be escaped
     * @see StringUtil#unescape(String) for converting back
     * @return escapen string
     */
    public static String escape(final String str) {
        final StringBuilder res = new StringBuilder(str.length() << 1);
        for (int i = 0; i < str.length(); i++) {
            final char c = str.charAt(i);
            switch (c) {
            case '\'':
                res.append("\\\'");
                break;
            case '\"':
                res.append("\\\"");
                break;
            case '\\':
                res.append("\\\\");
                break;
            case '\r':
                res.append("\\r");
                break;
            case '\n':
                res.append("\\n");
                break;
            case '\t':
                res.append("\\t");
                break;
            case '\f':
                res.append("\\f");
                break;
            default:
                res.append(c);
            }
        }
        return res.toString();
    }

    /**
     * Unescape special characters (CR, LF, Tab, FF) encoded accroding to
     * Java(C) conventions. It decodes backslashed characters simply by
     * interpreting next characted literally. For example, if you unescape <code>This\ris\na\t\\string\"\'\a\b\c\d</code>
     * you will get <code>ThisCRisLFaTAB\string"'abcd</code>
     *
     * @see StringUtil#escape(String) for escaping strings
     */
    public static String unescape(final String str) {
        final StringBuilder res = new StringBuilder(str.length());
        boolean slashFound = false;
        for (int i = 0; i < str.length(); i++) {
            final char c = str.charAt(i);
            if (slashFound) {
                switch (c) {
                case 'n':
                    res.append('\n');
                    break;
                case 'r':
                    res.append('\r');
                    break;
                case 't':
                    res.append('\t');
                    break;
                case 'f':
                    res.append('\f');
                    break;
                default:
                    res.append(c);
                    break;
                }
                slashFound = false;
            } else if (c == '\\') {
                slashFound = true;
            } else
                res.append(c);
        }
        return res.toString();
    }

    /**
     * Encode string as if we wanted to use it in source code (means escape it
     * and put quotes around it). So, for ex. if you encode <code>ThisCRisLFaTAB\string"'</code>
     * You will get <code>"This\ris\na\t\\string\"\'"</code>
     *
     * @see StringUtil#unstringalize(String) for converting back and
     *      StringUtil#escape(String) for escaping strings without putting them into
     *      quotes.
     */
    public static String stringalize(final String str) {
        if (str == null)
            return "null";
        else
            return '\"' + escape(str) + '\"';
    }

    /**
     * Decode string in source code format by unescaping it and removing quotes
     * around it. Recognizes single quotes and double quotes (Java recognizes
     * only double quotes). For ex. if you decode <code>"This\ris\na\t\\string\"\'\a\b\c\d"</code> You will get <code>ThisCRisLFaTab\string"'abcd</code>
     *
     * @throws StringFormatException
     *             when str is not enclosed in quotes properly
     */
    public static String unstringalize(final String str)
            throws StringFormatException {
        if (str.length() < 2)
            throw new StringFormatException(str, "String too small");
        if (str.charAt(0) != '\"')
            throw new StringFormatException(str, "Missing left parentize");
        if (str.charAt(str.length() - 1) != '\"')
            throw new StringFormatException(str, "Missing right parentize");
        return unescape(str.substring(1, str.length() - 1));
    }
    
    /** @return true if character id space, tab or form-feed character */
    public static boolean isWhitespace(final char c) {
    	return c == ' ' || c == '\t' || c == '\f';
    }

    public static final Charset utf8Charset = Charset.forName("UTF-8"),
                                asciiCharset = Charset.forName("ASCII");
    
    /** Convert string to utf-8 bytes sequence */
    public static byte [] toUTF8(final String s) {
    	return s.getBytes(utf8Charset);
    }
    
    /** Convert string to ascii bytes sequence */
    public static byte [] toASCII (final String s) {
    	return s.getBytes(asciiCharset);
    }

    /** @return true if string is either empty or null */
	public static boolean isEmpty(final String str) {
		return str == null || str.length() == 0;
	} 
	
    /**
     * Encode string into apllication/www-urlencoded format
     *
     * See {@link Util#urlencode(String)} for decoding strings
     */
    public static String urlencode(final String val) {
        try {
            return URLEncoder.encode(val, "UTF-8");
        } catch (final UnsupportedEncodingException e) {
            throw new FatalError(e);
        }
    }

    /**
     * Decode string from application/www-urlencoded format
     *
     * See {@link Util#urldecode(String)} for encoding strings
     */
    public static String urldecode(final String encoded) {
    	return urldecode(encoded, "UTF-8");
    }

    /** Decode string from application/www-urlencoded format in given charset */
    public static String urldecode(final String encoded, final String charset) {
        try {
            return URLDecoder.decode(encoded, charset);
        } catch (final UnsupportedEncodingException e) {
            throw new FatalError(e);
        }
    }
    
    /** @see StringUtil#urlParamsParse(String, String) */
    public static Map<String, List<String>> urlParamsParse(final String source)
    {
    	return urlParamsParse(source, "UTF-8");
    }

    /**
     * Parse parameters in apllication/www-urlencoded format. Note that you
     * should never pass whole url to this function. You should pass parameters
     * part (that comes after '?') only, so <code>
     * urlParamsParse("http://www.my.com/cgi?param=v1&param2=v2"); // Wrong!
     * urlParamsParse("param=v1&param2=v2"); //Right
     * </code>
     *
     * Since there can be multiple values for each parameter, this funciton
     * returns List&lt;String&gt; for every parameter. To simplify this, please
     * consider {@link GetParameters} class
     */
    public static Map<String, List<String>> urlParamsParse(
    		final String source, final String charset
    ){
        assert !source.startsWith("http://") :
        	"You should pass parameters part (after '?') of the query !";
        final Map<String, List<String>> res =
        	new HashMap<String, List<String>>();
        if (source.length()==0) return res;
        for (String part : split(source, '&')) {
            // Split name/val
            final String[] nameval = split(part, '=', 1);
            final String name = urldecode(nameval[0], charset),
            	         value;
            if (nameval.length == 1)
                value = "";
            else
            	value = urldecode(nameval[1], charset);

            // Get name bag or create new
            final List<String> bag;
            if (res.containsKey(name))
                bag = res.get(name);
            else {
                bag = new ArrayList<String>();
                res.put(name, bag);
            }

            // Add value to bag
            bag.add(value);
        }
        return res;
    }
    
    public static void dumpParameterIfNotEmpty(final StringBuilder res, Object value, final String name) {
		if (value == null) return;
		final String v;
		if (value instanceof String) {
			final String str = (String) value;
			if (str.length()==0) return;
			v = StringUtil.stringalize(str);
		} else {
			v = value.toString();
		}
		
		res.append(name);
		res.append('=');
		res.append(v);
		res.append(' ');
	}

	public static String strip(final String string) {
		final int len = string.length();
		int start = 0;
		while (start < len && isStopSymbol(string.charAt(start))) start++;
		int end = len-1;
		while (end>=0 && isStopSymbol(string.charAt(end))) end--;
		if (end<=start) return "";
		return string.substring(start, end+1);
	}
	
	private static boolean isStopSymbol(final char c) {
		return c == ' ' || c == '\n' || c == '\t' || c == '\f' || c == '\r';
	}

	/**
	 * "Fit" string into fixed maxLength. Will replace string tail with ellipsis '...'
	 * @param maxLength must be &gt; 0
	 * @throws IllegalArgumentException when maxLength &lt; 0
	 */
	public static String ellipsis(final String string, int maxLength) {
		if (maxLength<0) throw new IllegalArgumentException ("maxLength must be >0, got="+maxLength);
		if (maxLength > string.length()) return string;
		
		int trimTo = maxLength-3;
		if (trimTo < 0) trimTo = 0;
		
		final StringBuilder res = new StringBuilder(string);
		res.setLength(trimTo);
		while (res.length()<maxLength) res.append('.');
		
		return res.toString();
	}

	public static final int BEAUTIFY_CAPS_HINTS_RUSSIAN = 1;
	
	private static Map<String, String> russianAbbreviations;
	
	private static Map<String, String> getRussianAbbreviations() {
		if (russianAbbreviations != null) return russianAbbreviations;
		final Map<String, String> set = new HashMap<String, String>();
		final String path = '/'+StringUtil.class.getCanonicalName().replace('.', '/')+".russianAbbreviations.res";
		final InputStream in = StringUtil.class.getResourceAsStream(path);
		if (in != null) {
			try {
				final BufferedReader r = new BufferedReader(new InputStreamReader(in, "UTF-8"));
				String line = r.readLine(); 
				while (line != null) {
					set.put("(^|\\s)"+capitalizeFirst(line.toLowerCase()+"(\\s|$)"), "$1"+line+"$2");
					line = r.readLine();
				}
			} catch (final IOException e) {
				throw new FatalError(e);
			} finally {
				try {
					in.close();
				} catch (final IOException e) {
					throw new FatalError(e);
				}
			}
		}
		russianAbbreviations = set;
		return russianAbbreviations;
	}
	
	public static String beautifyCaps(String string) {
		return beautifyCaps(string, 0);
	}
	
	private static Pattern BEAUTIFY_CAPS;
	
	private static Pattern getBeautifyPattern() {
		if (BEAUTIFY_CAPS != null) return BEAUTIFY_CAPS;
		BEAUTIFY_CAPS = Pattern.compile("(\\S)(\\S*)");
		return BEAUTIFY_CAPS;
	}
	
	public static String beautifyCaps(String string, final int flags) {
		if (string.length()==0) return string;

		string = makeAllWordsStartWithUppercaseLetter(string);
		
		if ((flags & BEAUTIFY_CAPS_HINTS_RUSSIAN) != 0) {
			for (final Map.Entry<String, String> e : getRussianAbbreviations().entrySet()) {
				string = string.replaceAll(e.getKey(), e.getValue());
			}
		}
		return string;
	}
	
	public static String removeAbbreviations(String string) {
		for (final Map.Entry<String, String> e : getRussianAbbreviations().entrySet()) {
			final Pattern pat = Pattern.compile(e.getKey(), Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE);
			string = pat.matcher(string).replaceAll("$1$2");
		}
		return string;
	}
	
	private static String makeAllWordsStartWithUppercaseLetter(final String string) {
		Matcher m = getBeautifyPattern().matcher(string);
		final StringBuilder b = new StringBuilder();
		int pos = 0;
		while (m.find()) {
			final int start = m.start();
			final int end = m.end();
			b.append(string.substring(pos, start));
			b.append(m.group(1).toUpperCase());
			b.append(m.group(2).toLowerCase());
			pos = end;
		}
		return b.toString();
	}
	
	public static String capitalizeFirst(final String string) {
		if (string.length()==0) return string;
		return Character.toUpperCase(string.charAt(0))+string.substring(1);
	}

	/**
	 * @return true if string starts with prefix
	 */
	public static boolean startsWith(final String string, final String prefix) {
		if (string.length()<prefix.length()) return false;
		return string.regionMatches(0, prefix, 0, prefix.length());
	}
}
