/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.util.commutil;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import junit.framework.Assert;

import org.builtu.testing.BasicTestCase;
import org.builtu.util.commutil.StringUtil;
import org.builtu.util.commutil.Util;
import org.junit.Test;


public class TestUtilUrlDecode extends BasicTestCase {
    
	@Test public void testEncodeSimple() {
        final String source = " `1234567890-=\\qwertyuiop[]asdfghjkl;'zxcvbnm,./~!@#$%^&*()_+|QWERTYUIOP{}ASDFGHJKL:\"ZXCVBNM<>?";
        final String encoded = StringUtil.urlencode(source);
        assertEquals(
                "+%601234567890-%3D%5Cqwertyuiop%5B%5Dasdfghjkl%3B%27zxcvbnm%2C.%2F%7E%21%40%23%24%25%5E%26*%28%29_%2B%7CQWERTYUIOP%7B%7DASDFGHJKL%3A%22ZXCVBNM%3C%3E%3F",
                encoded);
        final String decoded = StringUtil.urldecode(encoded);
        assertEquals(source, decoded);
    }

    @Test public void testEncodeOnce() {
        final String source = rndAsciiString();
        final String encoded = StringUtil.urlencode(source);
        final String decoded = StringUtil.urldecode(encoded);
        assertEquals(source, decoded);
    }

    @Test public void testEncodeDecodeRandom() {
        final String source = rndAsciiString(4, 4);
        String encoded = source;
        final int passes = rnd(100);
        for (int i = 0; i < passes; i++)
            encoded = StringUtil.urlencode(encoded);
        for (int i = 0; i < passes; i++)
            encoded = StringUtil.urldecode(encoded);

        assertEquals(source, encoded);
    }

    @Test public void testParseSimple() {
        final String source = "name1=val1&name2=val2";
        final Map<String, List<String>> parsed = StringUtil.urlParamsParse(source);
        assertEquals("val1", parsed.get("name1").get(0));
        assertEquals("val2", parsed.get("name2").get(0));
        assertEquals(null, parsed.get("not_exists"));
    }

    @Test public void testParseSpecial() {
        final String source = StringUtil.urlencode("name1~!@#$%^&*()_") + "="
                + StringUtil.urlencode("val1~!@#$%^&*()") + "&name2=val2";
        final Map<String, List<String>> parsed = StringUtil.urlParamsParse(source);
        assertEquals("val1~!@#$%^&*()", parsed.get("name1~!@#$%^&*()_").get(0));
        assertEquals("val2", parsed.get("name2").get(0));
        assertEquals(null, parsed.get("not_exists"));
    }

    @Test public void testParseRandom() {
        setSeed(1206696499609L);

        final Map<String, List<String>> source = buildSource();

        final String reqString = assembleParams(source);

        final Map<String, List<String>> parsed = StringUtil.urlParamsParse(reqString);

        assertEquals(source.keySet(), parsed.keySet());
        for (String name : source.keySet()) {
            final List<String> expected = source.get(name), actual = parsed
                    .get(name);
            assertEquals("mismatch at '" + name + "'", expected, actual);
        }
        assertEquals(source, parsed);
    }
    
    @Test public void testPlus() {
    	Assert.assertEquals(" ", Util.urldecode("+"));
    }

    private Map<String, List<String>> buildSource() {
        final Map<String, List<String>> res = new HashMap<String, List<String>>();
        final int total_params = 1 + rnd(3);

        for (int i = 0; i < total_params; i++) {
            final String name = rndPrintableString(0, 5);
            if (res.containsKey(name))
                continue;

            final int total_values = 1 + rnd(5);
            final List<String> vals = new ArrayList<String>();
            for (int j = 0; j < total_values; j++)
                vals.add(rndPrintableString(0, 5));
            res.put(name, vals);
        }
        return res;
    }

    private String assembleParams(final Map<String, List<String>> params) {
        final StringBuilder res = new StringBuilder();
        for (String name : params.keySet()) {
            for (String val : params.get(name)) {
                if (res.length() > 0)
                    res.append('&');
                res.append(StringUtil.urlencode(name));
                res.append('=');
                res.append(StringUtil.urlencode(val));
            }
        }
        return res.toString();
    }
}
