/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.util.fs.transactions;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.channels.FileLock;

/**
 * Exclusive lock-on file mechanism. Uses nio channels to acquire actual lock.
 * Simple usage:
 * <pre>
 * final FileExclusiveLock lock = new FileExclusiveLock("mylock.lck");
 * if (lock.acquire()) {
 *   try {
 *     process(lock.getFile());
 *   } finally {
 *     lock.release(); 
 *   }
 * }
 * </pre>
 * Or, little bit simplier:
 * <pre>
 * final FileExclusiveLock lock = new FileExclusiveLock("mylock.lck");
 * try {
 *   if (lock.acquire()) process(lock.getFile());
 * } finally {
 *   lock.release(); 
 * }
 * </pre>
 * @author borisman
 *
 */
public class FileExclusiveLock {
	private final String lockPath, mode;
	private RandomAccessFile file;
    private FileLock lock;

    public FileExclusiveLock(final String lockPath) throws IOException {
		this(lockPath, "rws");
	}
    
    public FileExclusiveLock(final String lockPath, final String mode) throws IOException {
    	this.lockPath = lockPath;
    	this.mode = mode;
    }
    
    public boolean acquire() {
    	try {
            file = new RandomAccessFile(lockPath, mode);
            lock = file.getChannel().tryLock();
            if (lock==null) {
            	file.close();
            	file = null;
            	return false;
            }
            //file.setLength(0);
            //file.writeBytes(String.valueOf(new Date()));
    	} catch (final Throwable t) {
    		if (lock != null) release();
    	}
        return true;
    }
    
    public void release() {
        if (lock != null) {
        	try { lock.release(); } catch (final Throwable t){};
        	lock =  null;
        }
    	if (file != null) {
    		try { file.close(); } catch (final Throwable t){};
    		file = null;
    	}
    }
    
    public boolean isAcquired() {
    	return lock!=null;
    }
    
    public RandomAccessFile getFile() {
    	return file;
    }
    
    public String getPath() {
    	return lockPath;
    }
}
