package org.builtu.webutil.ordreq;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;

import org.apache.log4j.Logger;
import org.builtu.exceptions.FatalError;
import org.builtu.util.commutil.StringUtil;


public class OrderSessionRequestsFilter implements Filter {
	
	private static final Logger log = Logger.getLogger(OrderSessionRequestsFilter.class);
	
	private static class BehaviorMappping {
		public final String prefix;
		public final LockedBehavior behavior;
		
		public BehaviorMappping(final String mappingEntry) {
			final List<String> p = StringUtil.rsplit_list(mappingEntry, '=', 1);
			if (p.size()!=2) throw new FatalError("Bad mapping in policy: '"+mappingEntry+'\'');
			this.prefix = p.get(0).trim();
			this.behavior = LockedBehavior.valueOf(p.get(1).trim());
		}
	}
	
	private final List<BehaviorMappping> behaviors = new ArrayList<BehaviorMappping>();

	@Override public void init(final FilterConfig config) throws ServletException {
		final String policy = config.getInitParameter("policy");
		if (policy==null) {
			log.info("init, defaut polict is LOCK all requests by session");
			return;
		}
		
		for (final String s : StringUtil.split_list(policy, ',')) {
			behaviors.add(new BehaviorMappping(s.trim()));
		}
		
		final StringBuilder message = new StringBuilder(policy.length()+40);
		message.append("init policy:");
		final int len = message.length();
		for (final BehaviorMappping m : behaviors) {
			if (message.length()>len) message.append(", ");
			message.append(m.prefix);
			message.append('=');
			message.append(m.behavior);
		}
		
		log.info(message);
		
	}

	@Override public void destroy() {
		log.info("destroy");
	}
	
	private final OrdreqLock lock = new OrdreqLock();

	@Override public void doFilter(
			final ServletRequest req, 
			final ServletResponse resp,
			final FilterChain chain
	) throws IOException, ServletException {
		if (!(req instanceof HttpServletRequest)) {
			chain.doFilter(req, resp);
			return;
		}
		final HttpServletRequest hReq = (HttpServletRequest) req;
		final String sessionId = hReq.getSession().getId();

		if (sessionId == null) {
			chain.doFilter(req, resp);
			return;
		}
		
		final String path = hReq.getServletPath()+hReq.getPathInfo();
		final LockedBehavior beh = resolveBehavior(path);
		
		if (beh == LockedBehavior.SKIP && lock.wouldBlock(sessionId)) {
			log.debug("behavior is "+beh+", and request would block, skipping");
			chain.doFilter(req, resp);
			return;
		}
		
		lock.lock(sessionId);
		try {
			chain.doFilter(req, resp);
		} finally {
			lock.unlock(sessionId);
		}
		
	}

	private LockedBehavior resolveBehavior(final String path) {
		for (final BehaviorMappping m : behaviors) {
			if (path.startsWith(m.prefix)) return m.behavior;
		}
		return LockedBehavior.LOCK;
	}


}
